@@ -21,6 +21,7 @@ import ants
 import antspynet
 import antspyt1w
 import tensorflow as tf
+from tensorflow.python.eager.context import eager_mode, graph_mode
 
 from multiprocessing import Pool
 
@@ -292,10 +293,10 @@ def dbpn(input_image_size,
 
 # generate a random corner index for a patch
 
-def get_random_base_ind( full_dims, off = 10, patchWidth = 96 ):
+def get_random_base_ind( full_dims, patchWidth, off=8 ):
     baseInd = [None,None,None]
     for k in range(3):
-        baseInd[k]=random.sample( range( off, full_dims[k]-1-patchWidth ), 1 )[0]
+        baseInd[k]=random.sample( range( off, full_dims[k]-1-patchWidth[k] ), 1 )[0]
     return baseInd
 
 
@@ -303,7 +304,7 @@ def get_random_base_ind( full_dims, off = 10, patchWidth = 96 ):
 def get_random_patch( img, patchWidth ):
     mystd = 0
     while mystd == 0:
-        inds = get_random_base_ind( full_dims = img.shape )
+        inds = get_random_base_ind( full_dims = img.shape, patchWidth=patchWidth, off=8 )
         hinds = [None,None,None]
         for k in range(len(inds)):
             hinds[k] = inds[k] + patchWidth[k]
@@ -315,7 +316,7 @@ def get_random_patch_pair( img, img2, patchWidth ):
     mystd = mystd2 = 0
     ct = 0
     while mystd == 0 or mystd2 == 0:
-        inds = get_random_base_ind( full_dims = img.shape )
+        inds = get_random_base_ind( full_dims = img.shape, patchWidth=patchWidth, off=8  )
         hinds = [None,None,None]
         for k in range(len(inds)):
             hinds[k] = inds[k] + patchWidth[k]
@@ -332,7 +333,7 @@ def pseudo_3d_vgg_features( inshape = [128,128,128], layer = 4, angle=0, pretrai
     def getLayerScaleFactorForTransferLearning( k, w3d, w2d ):
         myfact = np.round( np.prod( w3d[k].shape ) / np.prod(  w2d[k].shape) )
         return myfact
-    vgg19 = tf.keras.applications.VGG19( 
+    vgg19 = tf.keras.applications.VGG19(
             include_top = False, weights = "imagenet",
             input_shape = [inshape[0],inshape[1],3],
             classes = 1000 )
@@ -345,7 +346,7 @@ def pseudo_3d_vgg_features( inshape = [128,128,128], layer = 4, angle=0, pretrai
     vggmodelRaw = antspynet.create_vgg_model_3d(
             [inshape[0],inshape[1],inshape[2],1],
             number_of_classification_labels = 1000,
-            layers = [1, 2, 3, 4, 4], 
+            layers = [1, 2, 3, 4, 4],
             lowest_resolution = 64,
             convolution_kernel_size= (3, 3, 3), pool_size = (2, 2, 2),
             strides = (2, 2, 2), number_of_dense_units= 4096, dropout_rate = 0,
@@ -353,12 +354,12 @@ def pseudo_3d_vgg_features( inshape = [128,128,128], layer = 4, angle=0, pretrai
     if verbose:
         print( vggmodelRaw.layers[layer_index] )
         print( vggmodelRaw.layers[layer_index].name )
-        print( vgg19.layers[layer_index] ) 
-        print( vgg19.layers[layer_index].name ) 
-    feature_extractor_2d = tf.keras.Model( 
+        print( vgg19.layers[layer_index] )
+        print( vgg19.layers[layer_index].name )
+    feature_extractor_2d = tf.keras.Model(
             inputs = vgg19.input,
             outputs = vgg19.layers[layer_index].output)
-    feature_extractor = tf.keras.Model( 
+    feature_extractor = tf.keras.Model(
             inputs = vggmodelRaw.input,
             outputs = vggmodelRaw.layers[layer_index].output)
     wts_2d = feature_extractor_2d.weights
@@ -367,7 +368,7 @@ def pseudo_3d_vgg_features( inshape = [128,128,128], layer = 4, angle=0, pretrai
         if len(a.shape) != len(b.shape):
                 return False
         for j in range(len(a.shape)):
-            if a.shape[j] != b.shape[j]: 
+            if a.shape[j] != b.shape[j]:
                 return False
         return True
     for ww in range(len(wts)):
@@ -432,25 +433,59 @@ def get_grader_feature_network( layer=6 ):
     return tf.keras.Model( inputs=grader.inputs, outputs=grader.layers[layer].output )
 
 
-def default_dbpn( 
+def default_dbpn(
     strider, # length should equal dimensionality
     dimensionality = 3,
     nfilt=64,
     nff = 256,
     convn = 6,
     lastconv = 3,
-    nbp=7, 
+    nbp=7,
     nChannelsIn=1,
     nChannelsOut=1,
-    sigmoid_second_channel=False
+    option = None,
+    intensity_model=None,
+    segmentation_model=None,
+    sigmoid_second_channel=False,
+    clone_intensity_to_segmentation=False,
+    pro_seg = 0,
+    freeze = False,
+    verbose=False
  ):
+    if option == 'tiny':
+        nfilt=32
+        nff = 64
+        convn = 3
+        lastconv = 1
+        nbp=2
+    elif option == 'small':
+        nfilt=32
+        nff = 64
+        convn = 6
+        lastconv = 3
+        nbp=4
+    elif option == 'medium':
+        nfilt=64
+        nff = 128
+        convn = 6
+        lastconv = 3
+        nbp=4
+    else:
+        option='large'
+    if verbose:
+        print("Building mode of size: " + option)
+        if intensity_model is not None:
+            print("user-passed intensity model will be frozen - only segmentation will train")
+        if segmentation_model is not None:
+            print("user-passed segmentation model will be frozen - only intensity will train")
+
     if len(strider) != dimensionality:
         raise Exception("len(strider) != dimensionality")
     # **model instantiation**: these are close to defaults for the 2x network.<br>
     # empirical evidence suggests that making covolutions and strides evenly<br>
     # divisible by each other reduces artifacts.  2*3=6.
     # ofn='./models/dsr3d_'+str(strider)+'up_' + str(nfilt) + '_' + str( nff ) + '_' + str(convn)+ '_' + str(lastconv)+ '_' + str(os.environ['CUDA_VISIBLE_DEVICES'])+'_v0.0.h5'
-    if dimensionality == 2:
+    if dimensionality == 2 :
         mdl = dbpn( (None,None,nChannelsIn),
             number_of_outputs=nChannelsOut,
             number_of_base_filters=nfilt,
@@ -458,10 +493,10 @@ def default_dbpn(
             number_of_back_projection_stages=nbp,
             convolution_kernel_size=(convn, convn),
             strides=(strider[0], strider[1]),
-            last_convolution=(lastconv, lastconv), 
-            number_of_loss_functions=1, 
+            last_convolution=(lastconv, lastconv),
+            number_of_loss_functions=1,
             interpolation='nearest')
-    if dimensionality == 3:
+    if dimensionality == 3 :
         mdl = dbpn( (None,None,None,nChannelsIn),
             number_of_outputs=nChannelsOut,
             number_of_base_filters=nfilt,
@@ -470,22 +505,142 @@ def default_dbpn(
             convolution_kernel_size=(convn, convn, convn),
             strides=(strider[0], strider[1], strider[2]),
             last_convolution=(lastconv, lastconv, lastconv), number_of_loss_functions=1, interpolation='nearest')
-    if sigmoid_second_channel:
-        mdlout = tf.split( mdl.outputs[0], 2, dimensionality+1)
-        mdlout[1] = tf.nn.sigmoid( mdlout [1] )
-        mdl.outputs[0] = tf.concat( mdlout, axis=dimensionality+1 )
-        mdl = Model(inputs=mdl.inputs, outputs=mdl.outputs )
+    if sigmoid_second_channel and pro_seg != 0 :
+        if dimensionality == 2 :
+            input_image_size = (None,None,2)
+            if intensity_model is None:
+                intensity_model = dbpn( (None,None,1),
+                    number_of_outputs=1,
+                    number_of_base_filters=nfilt,
+                    number_of_feature_filters=nff,
+                    number_of_back_projection_stages=nbp,
+                    convolution_kernel_size=(convn, convn),
+                    strides=(strider[0], strider[1]),
+                    last_convolution=(lastconv, lastconv),
+                    number_of_loss_functions=1,
+                    interpolation='nearest')
+            else:
+                if freeze:
+                    for layer in intensity_model.layers:
+                        layer.trainable = False
+            if segmentation_model is None:
+                segmentation_model = dbpn( (None,None,1),
+                        number_of_outputs=1,
+                        number_of_base_filters=nfilt,
+                        number_of_feature_filters=nff,
+                        number_of_back_projection_stages=nbp,
+                        convolution_kernel_size=(convn, convn),
+                        strides=(strider[0], strider[1]),
+                        last_convolution=(lastconv, lastconv),
+                        number_of_loss_functions=1, interpolation='linear')
+            else:
+                if freeze:
+                    for layer in segmentation_model.layers:
+                        layer.trainable = False
+        if dimensionality == 3 :
+            input_image_size = (None,None,None,2)
+            if intensity_model is None:
+                intensity_model = dbpn( (None,None,None,1),
+                    number_of_outputs=1,
+                    number_of_base_filters=nfilt,
+                    number_of_feature_filters=nff,
+                    number_of_back_projection_stages=nbp,
+                    convolution_kernel_size=(convn, convn, convn),
+                    strides=(strider[0], strider[1], strider[2]),
+                    last_convolution=(lastconv, lastconv, lastconv),
+                    number_of_loss_functions=1, interpolation='nearest')
+            else:
+                if freeze:
+                    for layer in intensity_model.layers:
+                        layer.trainable = False
+            if segmentation_model is None:
+                segmentation_model = dbpn( (None,None,None,1),
+                        number_of_outputs=1,
+                        number_of_base_filters=nfilt,
+                        number_of_feature_filters=nff,
+                        number_of_back_projection_stages=nbp,
+                        convolution_kernel_size=(convn, convn, convn),
+                        strides=(strider[0], strider[1], strider[2]),
+                        last_convolution=(lastconv, lastconv, lastconv),
+                        number_of_loss_functions=1, interpolation='linear')
+            else:
+                if freeze:
+                    for layer in segmentation_model.layers:
+                        layer.trainable = False
+        if verbose:
+            print( "len intensity_model layers : " + str( len( intensity_model.layers )))
+            print( "len intensity_model weights : " + str( len( intensity_model.weights )))
+            print( "len segmentation_model layers : " + str( len( segmentation_model.layers )))
+            print( "len segmentation_model weights : " + str( len( segmentation_model.weights )))
+        if clone_intensity_to_segmentation:
+            for k in range(len( segmentation_model.weights )):
+                if k < len( intensity_model.weights ):
+                    if intensity_model.weights[k].shape == segmentation_model.weights[k].shape:
+                        segmentation_model.weights[k] = intensity_model.weights[k]
+        inputs = tf.keras.Input(shape=input_image_size)
+        insplit = tf.split( inputs, 2, dimensionality+1)
+        outputs = [
+            intensity_model( insplit[0] ),
+            tf.nn.sigmoid( segmentation_model( insplit[1] ) ) ]
+        mdlout = tf.concat( outputs, axis=dimensionality+1 )
+        return Model(inputs=inputs, outputs=mdlout )
+    if pro_seg > 0 and intensity_model is not None:
+        if verbose and freeze:
+            print("Add a segmentation arm to the end. freeze intensity. intensity_model(seg) => conv => sigmoid")
+        if verbose and not freeze:
+            print("Add a segmentation arm to the end. freeze intensity. intensity_model(seg) => conv => sigmoid")
+        if freeze:
+            for layer in intensity_model.layers:
+                layer.trainable = False
+        if dimensionality == 2 :
+            input_image_size = (None,None,2)
+        elif dimensionality == 3 :
+            input_image_size = (None, None,None,2)
+        if dimensionality == 2:
+            myconv = Conv2D
+            firstConv = (convn,convn)
+            firstStrides=(1,1)
+            smashConv=(pro_seg,pro_seg)
+        if dimensionality == 3:
+            myconv = Conv3D
+            firstConv = (convn,convn,convn)
+            firstStrides=(1,1,1)
+            smashConv=(pro_seg,pro_seg,pro_seg)
+        inputs = tf.keras.Input(shape=input_image_size)
+        insplit = tf.split( inputs, 2, dimensionality+1)
+        # define segmentation arm
+        seggit = intensity_model( insplit[1] )
+        L0 = myconv(filters=nff,
+                    kernel_size=firstConv,
+                    strides=firstStrides,
+                    kernel_initializer='glorot_uniform',
+                    padding='same')(seggit)
+        L1 = myconv(filters=nff,
+                    kernel_size=firstConv,
+                    strides=firstStrides,
+                    kernel_initializer='glorot_uniform',
+                    padding='same')(L0)
+        L2 = myconv(filters=1,
+                    kernel_size=smashConv,
+                    strides=firstStrides,
+                    kernel_initializer='glorot_uniform',
+                    padding='same')(L1)
+        outputs = [
+            intensity_model( insplit[0] ),
+            tf.nn.sigmoid( L2 ) ]
+        mdlout = tf.concat( outputs, axis=dimensionality+1 )
+        return Model(inputs=inputs, outputs=mdlout )
     return mdl
 
-def image_patch_training_data_from_filenames( 
+def image_patch_training_data_from_filenames(
     filenames,
     target_patch_size,
     target_patch_size_low,
-    nPatches = 128, 
+    nPatches = 128,
     istest   = False,
-    patch_scaler=True, 
+    patch_scaler=True,
     to_tensorflow = False,
-    verbose = False 
+    verbose = False
     ):
     # **data generation**<br>
     # recent versions of tensorflow/keras allow data generators to be passed<br>
@@ -510,7 +665,7 @@ def image_patch_training_data_from_filenames(
     patchesResam = np.zeros(shape=shaperlo)
     patchesUp = None
     if istest:
-        patchesUp = np.zeros(shape=patchesOrig.shape)    
+        patchesUp = np.zeros(shape=patchesOrig.shape)
     for myn in range(nPatches):
             if verbose:
                 print(myn)
@@ -558,15 +713,15 @@ def image_patch_training_data_from_filenames(
     return patchesResam, patchesOrig, patchesUp
 
 
-def seg_patch_training_data_from_filenames( 
+def seg_patch_training_data_from_filenames(
     filenames,
     target_patch_size,
     target_patch_size_low,
-    nPatches = 128, 
+    nPatches = 128,
     istest   = False,
-    patch_scaler=True, 
+    patch_scaler=True,
     to_tensorflow = False,
-    verbose = False 
+    verbose = False
     ):
     if verbose:
         print("begin seg_patch_training_data_from_filenames")
@@ -585,7 +740,7 @@ def seg_patch_training_data_from_filenames(
     patchesResam = np.zeros(shape=shaperlo)
     patchesUp = None
     if istest:
-        patchesUp = np.zeros(shape=patchesOrig.shape)    
+        patchesUp = np.zeros(shape=patchesOrig.shape)
     for myn in range(nPatches):
             if verbose:
                 print(myn)
@@ -723,22 +878,22 @@ def numpy_generator( filenames ):
     patchesResam=patchesOrig=patchesUp=None
     yield (patchesResam, patchesOrig,patchesUp)
 
-def image_generator( 
+def image_generator(
     filenames,
     nPatches,
     target_patch_size,
     target_patch_size_low,
-    patch_scaler=True, 
+    patch_scaler=True,
     istest=False,
     verbose = False ):
     while True:
-        patchesResam, patchesOrig, patchesUp = image_patch_training_data_from_filenames( 
+        patchesResam, patchesOrig, patchesUp = image_patch_training_data_from_filenames(
             filenames,
             target_patch_size = target_patch_size,
             target_patch_size_low = target_patch_size_low,
-            nPatches = nPatches, 
+            nPatches = nPatches,
             istest   = istest,
-            patch_scaler=patch_scaler, 
+            patch_scaler=patch_scaler,
             to_tensorflow = True,
             verbose = verbose )
         if istest:
@@ -746,22 +901,22 @@ def image_generator(
         yield (patchesResam, patchesOrig)
 
 
-def seg_generator( 
+def seg_generator(
     filenames,
     nPatches,
     target_patch_size,
     target_patch_size_low,
-    patch_scaler=True, 
+    patch_scaler=True,
     istest=False,
     verbose = False ):
     while True:
-        patchesResam, patchesOrig, patchesUp = seg_patch_training_data_from_filenames( 
+        patchesResam, patchesOrig, patchesUp = seg_patch_training_data_from_filenames(
             filenames,
             target_patch_size = target_patch_size,
             target_patch_size_low = target_patch_size_low,
-            nPatches = nPatches, 
+            nPatches = nPatches,
             istest   = istest,
-            patch_scaler=patch_scaler, 
+            patch_scaler=patch_scaler,
             to_tensorflow = True,
             verbose = verbose )
         if istest:
@@ -769,22 +924,23 @@ def seg_generator(
         yield (patchesResam, patchesOrig)
 
 
-def train( 
-    mdl, 
-    filenames_train, 
-    filenames_test, 
+def train(
+    mdl,
+    filenames_train,
+    filenames_test,
     target_patch_size,
     target_patch_size_low,
     output_prefix,
     n_test = 8,
-    learning_rate=5e-5, 
-    feature_layer = 6, 
+    learning_rate=5e-5,
+    feature_layer = 6,
     feature = 2,
     tv = 0.1,
     max_iterations = 1000,
     batch_size = 1,
     save_all_best = False,
     feature_type = 'grader',
+    check_eval_data_iteration = 20,
     verbose = False  ):
     colnames = ['train_loss','test_loss','best','eval_psnr','eval_psnr_lin']
     training_path = np.zeros( [ max_iterations, len(colnames) ] )
@@ -794,15 +950,17 @@ def train(
     if feature_type == 'grader':
         feature_extractor = get_grader_feature_network( feature_layer )
     elif feature_type == 'vggrandom':
-        feature_extractor = pseudo_3d_vgg_features( target_patch_size, feature_layer, pretrained=False )
+        with eager_mode():
+            feature_extractor = pseudo_3d_vgg_features( target_patch_size, feature_layer, pretrained=False )
     elif feature_type == 'vgg':
-        feature_extractor = pseudo_3d_vgg_features_unbiased( target_patch_size, feature_layer )
+        with eager_mode():
+            feature_extractor = pseudo_3d_vgg_features_unbiased( target_patch_size, feature_layer )
     else:
         raise Exception("feature type does not exist")
     if verbose:
         print("begin train generator")
-    mydatgen = image_generator( 
-        filenames_train, 
+    mydatgen = image_generator(
+        filenames_train,
         nPatches=1,
         target_patch_size=target_patch_size,
         target_patch_size_low=target_patch_size_low,
@@ -845,8 +1003,9 @@ def train(
         if verbose:
             print( "preset weights:" )
     else:
-        wts = auto_weight_loss( mdl, feature_extractor, patchesResamTeTf, patchesOrigTeTf, 
-            feature=feature, tv=tv )
+        with eager_mode():
+            wts = auto_weight_loss( mdl, feature_extractor, patchesResamTeTf, patchesOrigTeTf,
+                feature=feature, tv=tv )
         for k in range(len(wts)):
             training_weights[0,k]=wts[k]
         pd.DataFrame(training_weights, columns = ["msq","feat","tv"] ).to_csv( wts_csv )
@@ -854,7 +1013,7 @@ def train(
             print( "automatic weights:" )
     if verbose:
         print( wts )
-    def my_loss_6( y_true, y_pred, msqwt = wts[0], fw = wts[1], tvwt = wts[2], mybs = batch_size ): 
+    def my_loss_6( y_true, y_pred, msqwt = wts[0], fw = wts[1], tvwt = wts[2], mybs = batch_size ):
         squared_difference = tf.square(y_true - y_pred)
         if len( y_true.shape ) == 5:
             tdim = 3
@@ -896,7 +1055,7 @@ def train(
         training_path[myrs,1]=tracker.history['val_loss'][0]
         training_path[myrs,2]=0
         print( "ntrain: " + str(myrs) + " loss " + str( tracker.history['loss'][0] ) + ' val-loss ' + str(tracker.history['val_loss'][0]), flush=True  )
-        if myrs % 20 == 0:
+        if myrs % check_eval_data_iteration == 0:
             with tf.device("/cpu:0"):
                 myofn = output_prefix + "_best_mdl.h5"
                 if save_all_best:
@@ -927,18 +1086,18 @@ def binary_dice_loss(y_true, y_pred):
   y_pred_f = K.flatten(y_pred)
   intersection = K.sum(y_true_f * y_pred_f)
   return -1 * (2 * intersection + smoothing_factor)/(K.sum(y_true_f) +
-        K.sum(y_pred_f) + smoothing_factor) 
+        K.sum(y_pred_f) + smoothing_factor)
 
 def train_seg(
-    mdl, 
-    filenames_train, 
-    filenames_test, 
+    mdl,
+    filenames_train,
+    filenames_test,
     target_patch_size,
     target_patch_size_low,
     output_prefix,
     n_test = 8,
-    learning_rate=5e-5, 
-    feature_layer = 6, 
+    learning_rate=5e-5,
+    feature_layer = 6,
     feature = 2,
     tv = 0.1,
     dice = 0.5,
@@ -946,6 +1105,7 @@ def train_seg(
     batch_size = 1,
     save_all_best = False,
     feature_type = 'grader',
+    check_eval_data_iteration = 20,
     verbose = False  ):
     colnames = ['train_loss','test_loss','best','eval_psnr','eval_psnr_lin','eval_msq','eval_dice']
     training_path = np.zeros( [ max_iterations, len(colnames) ] )
@@ -960,8 +1120,8 @@ def train_seg(
         feature_extractor = pseudo_3d_vgg_features_unbiased( target_patch_size, feature_layer  )
     if verbose:
         print("begin train generator")
-    mydatgen = seg_generator( 
-        filenames_train, 
+    mydatgen = seg_generator(
+        filenames_train,
         nPatches=1,
         target_patch_size=target_patch_size,
         target_patch_size_low=target_patch_size_low,
@@ -1004,7 +1164,7 @@ def train_seg(
         if verbose:
             print( "preset weights:" )
     else:
-        wts = auto_weight_loss_seg( mdl, feature_extractor, patchesResamTeTf, patchesOrigTeTf, 
+        wts = auto_weight_loss_seg( mdl, feature_extractor, patchesResamTeTf, patchesOrigTeTf,
             feature=feature, tv=tv, dice=dice )
         for k in range(len(wts)):
             training_weights[0,k]=wts[k]
@@ -1013,7 +1173,7 @@ def train_seg(
             print( "automatic weights:" )
     if verbose:
         print( wts )
-    def my_loss_6( y_true, y_pred, msqwt = wts[0], fw = wts[1], tvwt = wts[2], dicewt=wts[3], mybs = batch_size ): 
+    def my_loss_6( y_true, y_pred, msqwt = wts[0], fw = wts[1], tvwt = wts[2], dicewt=wts[3], mybs = batch_size ):
         if len( y_true.shape ) == 5:
             tdim = 3
             myax = [1,2,3,4]
@@ -1059,7 +1219,7 @@ def train_seg(
         training_path[myrs,1]=tracker.history['val_loss'][0]
         training_path[myrs,2]=0
         print( "ntrain: " + str(myrs) + " loss " + str( tracker.history['loss'][0] ) + ' val-loss ' + str(tracker.history['val_loss'][0]), flush=True  )
-        if myrs % 20 == 0:
+        if myrs % check_eval_data_iteration == 0:
             with tf.device("/cpu:0"):
                 myofn = output_prefix + "_best_mdl.h5"
                 if save_all_best:
@@ -1091,11 +1251,11 @@ def train_seg(
     training_path = pd.DataFrame(training_path, columns = colnames )
     return training_path
 
-def inference( 
-    image, 
+def inference(
+    image,
     mdl,
-    truncation = [0.001,0.999],
-    segmentation=None, 
+    truncation = None,
+    segmentation=None,
     target_range=[1,0],
     match_intensity=True,
     verbose=False):
@@ -1123,13 +1283,22 @@ def inference(
             testarrout = mdl( testarr )
             for k in range(2):
                 upFactor.append( int( testarrout.shape[k+1]/testarr.shape[k+1]  )  )
+
+        return antspyt1w.super_resolution_segmentation_per_label(
+                pimg,
+                segmentation,
+                upFactor,
+                mdl,
+                segmentation_numbers=mynp,
+                target_range=target_range,
+                max_lab_plus_one=True  )
         return antspyt1w.super_resolution_segmentation_per_label(
             pimg, segmentation, upFactor, mdl,
             segmentation_numbers=mynp,
-            dilation_amount=0,
+            dilation_amount=8,
             probability_images=None,
-            probability_labels=None, 
-            max_lab_plus_one=True, 
+            probability_labels=None,
+            max_lab_plus_one=True,
             target_range=target_range,
             match_intensity=match_intensity,
             verbose=verbose)
@@ -1198,11 +1367,68 @@ def simulate_image( shaper=[32,32,32], n_levels=10, multiply=False ):
         img = img + temp
     return img
 
+def optimize_upsampling_shape( spacing, modality='T1', roundit=False, verbose=False ):
+    minspc = min( list( spacing ) )
+    maxspc = max( list( spacing ) )
+    ratio = maxspc/minspc
+    if ratio == 1.0:
+        ratio = 0.5
+    roundratio = np.round( ratio )
+    tarshaperaw = []
+    tarshape = []
+    tarshaperound = []
+    for k in range( len( spacing ) ):
+        locrat = spacing[k]/minspc
+        newspc = spacing[k] * roundratio
+        tarshaperaw.append( locrat )
+        if modality == "NM":
+            if verbose:
+                print("Using minspacing: 0.25")
+            if newspc < 0.25 :
+                locrat = spacing[k]/0.25
+        elif modality == "DTI":
+            if verbose:
+                print("Using minspacing: 1.0")
+            if newspc < 1.0 :
+                locrat = spacing[k]/1.0
+        else: # assume T1
+            if verbose:
+                print("Using minspacing: 0.35")
+            if newspc < 0.35 :
+                locrat = spacing[k]/0.35
+        myint = int( locrat )
+        if ( myint == 0 ):
+            myint = 1
+        if myint == 5:
+            myint = 4
+        if ( myint > 6 ):
+            myint = 6
+        tarshape.append( str( myint ) )
+        tarshaperound.append( str( int(np.round( locrat )) ) )
+    if verbose:
+        print("before emendation:")
+        print( tarshaperaw )
+        print( tarshaperound )
+        print( tarshape )
+    allone = True
+    if roundit:
+        tarshape = tarshaperound
+    for k in range( len( tarshape ) ):
+        if tarshape[k] != "1":
+            allone=False
+    if allone:
+        tarshape = ["2","2","2"] # default
+    return "x".join(tarshape)
 
-def compare_models( model_filenames, img, verbose=False ):
+def compare_models( model_filenames, img, n_classes=3, identifier=None, verbose=False ):
     """
     generate a dataframe computing some basic intensity metrics PSNR and SSIM
+
+    NOTE: when evaluating a 2-channel (segmentation) model - the focus should
+    be on the segmentation arm (DICE) alone ... the intensity component can
+    be evaluated on its own separately.
     """
+    padding=4
     mydf = pd.DataFrame()
     for k in range( len( model_filenames ) ):
         srmdl, upshape = read_srmodel( model_filenames[k] )
@@ -1220,7 +1446,7 @@ def compare_models( model_filenames, img, verbose=False ):
         if verbose:
             print( dimg )
         if upshape[3] == 2:
-            seghigh = ants.threshold_image( img,"Otsu",2)
+            seghigh = ants.threshold_image( img,"Otsu",n_classes)
             seglow = ants.resample_image( seghigh, tarshape, use_voxels=False, interp_type=1 )
             dimgup=inference( dimg, srmdl, segmentation = seglow, verbose=False )
             dimgupseg = dimgup['super_resolution_segmentation']
@@ -1235,7 +1461,7 @@ def compare_models( model_filenames, img, verbose=False ):
         padder = []
         dimwarning=False
         for jj in range(img.dimension):
-            padder.append( -2 )
+            padder.append( padding )
             if img.shape[jj] != imgblock.shape[jj]:
                 dimwarning=True
         if dimwarning:
@@ -1250,11 +1476,18 @@ def compare_models( model_filenames, img, verbose=False ):
             print( "resampled img shape" )
             print( imgblock.shape )
         a=[]
+        imgshape = []
         for aa in range(len(upshape)):
             a.append( str(upshape[aa]) )
+            if aa < len(imgblock.shape):
+                imgshape.append( str( imgblock.shape[aa] ) )
+        if identifier is None:
+            identifier=temp
         mydict = {
+            "identifier":identifier,
+            "imgshape":"x".join(imgshape),
             "mdl": temp,
-            "shape":"x".join(a),
+            "mdlshape":"x".join(a),
             "PSNR.LIN": antspynet.psnr( imgblock, dimglin ),
             "PSNR.SR": antspynet.psnr( imgblock, dimgup ),
             "SSIM.LIN": antspynet.ssim( imgblock, dimglin ),
